# -*- coding: utf-8 -*-
"""Version information for tantrum."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__version__ = "1.0.2"
""":obj:`str`: Version of package."""
__url__ = "https://github.com/lifehackjim/tantrum"
""":obj:`str`: URL of package."""
__author__ = "Jim Olsen"
""":obj:`str`: Auth of package."""
__title__ = "tantrum"
""":obj:`str`: Title of package."""
__project__ = "tantrum"
""":obj:`str`: Name of package."""
__author_email__ = "jim@lifehack.com"
""":obj:`str`: Author email of package."""
__description__ = "Tanium API workflow encapsulation"
""":obj:`str`: Description of package."""
__license__ = "MIT"
""":obj:`str`: License of package."""
__copyright__ = "Copyright Jim Olsen 2019"
""":obj:`str`: Copyright of package."""
